/*
 * Copyright (c) 2002-2006 Atheros Communications, Inc.
 * All rights reserved.
 *
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/cache.h>
#include <linux/sysctl.h>
#include <linux/proc_fs.h>
#include "sys/queue.h"

#include <asm/uaccess.h>

#include <net80211/if_media.h>
#include <net80211/ieee80211_var.h>

#include "if_athvar.h"
#include "if_athioctl.h"
#include "ah_desc.h"
#include "dfs.h"
#include "dfs_ioctl.h"
#include "ar5212/ar5212reg.h"

static	u_int32_t _round(int32_t val);
static	void dfs_print_delayline(struct ath_softc *sc, struct dfs_delayline *dl);
static	void dfs_reset_alldelaylines(struct ath_softc *sc);
static	void dfs_reset_delayline(struct dfs_delayline *dl);
static	void dfs_reset_radarq(struct ath_softc *sc);
static	void  dfs_add_pulse(struct ath_softc *sc, struct dfs_filter *rf,
			   struct dfs_event *re, u_int32_t deltaT);
static  int dfs_bin_check(struct ath_softc *sc, struct dfs_filter *rf,
                            u_int32_t deltaT, u_int32_t dur);
//static	int dfs_bin_fixedpattern_check(struct ath_softc *sc, struct dfs_filter *rf);

static	void dfs_nol_addchan(struct ath_softc *sc, HAL_CHANNEL *chan);
static	int dfs_bin5_addpulse(struct ath_softc *sc, struct dfs_bin5radars *br,
			       struct dfs_event *re, u_int64_t thists);
static	int dfs_bin5_check(struct ath_softc *sc);
static	void dfs_reset_ar(struct ath_softc *sc);
static	void dfs_reset_arq(struct ath_softc *sc);
static	struct dfs_pulse *get_dfs_radars(struct ath_softc *sc,
					 u_int32_t dfsdomain, int *numradars,
					 struct dfs_bin5pulses **bin5pulses,
					 int * numb5radars,
					 HAL_PHYERR_PARAM *pe);
static	int dfs_set_thresholds(struct ath_softc *sc, const u_int32_t threshtype,
			       const u_int32_t value);
static	void dfs_dump_pulselog(struct dfs_pulseline *pl);

#include "dfs_pulses.h"

#define	DPRINTF(sc, _m, _fmt, ...) do {				\
	if (sc->sc_debug & (_m))				\
		printk(_fmt, __VA_ARGS__);			\
} while (0)

#define DFS_MIN(a,b) ((a)<(b)?(a):(b))
#define DFS_MAX(a,b) ((a)>(b)?(a):(b))
#define DFS_DIFF(a,b) (DFS_MAX(a,b) - DFS_MIN(a,b))
#define	IS_5312_2_X(ah) \
	((((ah)->ah_macVersion) == AR_SREV_VERSION_VENICE) && \
	 (((ah)->ah_macRev == 2) || (((ah)->ah_macRev) == 7)))
#define IS_5424(ah) \
	((((ah)->ah_macVersion) == AR_SREV_5424) || \
	((((ah)->ah_macVersion) == AR_SREV_5413) && \
	 ((ah)->ah_macRev) <=3))
#define IS_5413(ah) \
	((((ah)->ah_macVersion) == AR_SREV_5413) || IS_5424(ah))
#define IS_5416_PCI(ah) (((ah)->ah_macVersion) == AR_SREV_VERSION_OWL_PCI)
#define IS_5416_PCIE(ah) (((ah)->ah_macVersion) == AR_SREV_VERSION_OWL_PCIE)
#define IS_5416(ah) (IS_5416_PCI(ah) || IS_5416_PCIE(ah))

static int domainoverride=DFS_UNINIT_DOMAIN;
MODULE_PARM(domainoverride, "i");
MODULE_PARM_DESC(domainoverride, "Override dfs domain");
static int usenol=1;
MODULE_PARM(usenol, "i");
MODULE_PARM_DESC(usenol, "Override the use of channel NOL");

/*
 * Finds the radar state entry that matches the current channel
 */

static u_int32_t
_round(int32_t val)
{
	u_int32_t ival,rem;

	if (val < 0)
		return 0;
	ival = val/100;
	rem = val-(ival*100);
	if (rem <50)
		return ival;
	else
		return(ival+1);
}

static void
dfs_print_delayline(struct ath_softc *sc, struct dfs_delayline *dl)
{
	int i=0,index;
	struct dfs_delayelem *de;

	index = dl->dl_lastelem;
 	for (i=0; i<dl->dl_numelems; i++) {
		de = &dl->dl_elems[index];
		DPRINTF(sc, ATH_DEBUG_DFS2, 
			"Elem %d: ts = %u (0x%x) dur=%u\n",i,
		       de->de_time, de->de_time, de->de_dur);
		index = (index - 1)& DFS_MAX_DL_MASK;
	}
}

/*
 * Clear all delay lines for all filter types
 */
static void
dfs_reset_alldelaylines(struct ath_softc *sc)
{
        struct ath_dfs *dfs=sc->sc_dfs;
        struct dfs_filtertype *ft = NULL;
        struct dfs_filter *rf;
        struct dfs_delayline *dl;
        struct dfs_pulseline *pl = &dfs->pulses;
        int i,j;

        if (dfs == NULL) {
                DPRINTF(sc, ATH_DEBUG_DFS, "%s: sc_dfs is NULL\n", __func__);
                return;
        }
	/* reset the pulse log */
	pl->pl_firstelem = pl->pl_numelems = 0;
	pl->pl_lastelem = DFS_MAX_PULSE_BUFFER_MASK;

	/* we probably do not need to memset all the delay lines as done below
	 * it should be okay to reset just the pointers like it is done above for pulse log.
	 *  revisit this if you face  any problems passing the channel closing requirements 
	 */
        for (i=0; i<DFS_MAX_RADAR_TYPES; i++) {
                ft = dfs->dfs_radarf[i];
                for (j=0; j<ft->ft_numfilters; j++) {
                        rf = &(ft->ft_filters[j]);
                        dl = &(rf->rf_dl);
                        if(dl != NULL) {
                                memset(dl, 0, sizeof(struct dfs_delayline));
                                dl->dl_lastelem = (0xFFFFFFFF) & DFS_MAX_DL_MASK;
                        }
                        else
                                printk("dfs_reset_alldelaylines: DL is NULL !!!!!!!!!!!\n");
                }
        }
        for (i=0; i<dfs->dfs_rinfo.rn_numbin5radars; i++) {
                memset(&(dfs->dfs_b5radars[i].br_elems[0]), 0,
                       sizeof(struct dfs_bin5elem)*DFS_MAX_B5_SIZE);
                dfs->dfs_b5radars[i].br_firstelem = 0;
                dfs->dfs_b5radars[i].br_numelems = 0;
                dfs->dfs_b5radars[i].br_lastelem = (0xFFFFFFFF)&DFS_MAX_B5_MASK;
        }
}

/*
 * Clear only a single delay line
 */

static void
dfs_reset_delayline(struct dfs_delayline *dl)
{
	memset(&(dl->dl_elems[0]), 0, sizeof(dl->dl_elems));
	dl->dl_lastelem = (0xFFFFFFFF)&DFS_MAX_DL_MASK;
}
	
/*
 * Clear/Reset radar information for a single radar state/channel
 * Also clears all radar delay lines for all filter types
 */

#if 0
static void
dfs_reset_radarchannel(struct ath_softc *sc, struct dfs_state *rs)
{
	dfs_reset_alldelaylines(sc);
	rs->rs_numradarevents=0;
}
#endif

static void
dfs_reset_radarq(struct ath_softc *sc)
{
	struct ath_dfs *dfs=sc->sc_dfs;
	struct dfs_event *event;

	if (dfs == NULL) {
		DPRINTF(sc, ATH_DEBUG_DFS, "%s: sc_dfs is NULL\n", __func__);
		return;
	}
	ATH_DFSQ_LOCK(dfs);
	ATH_DFSEVENTQ_LOCK(dfs);
	while (!STAILQ_EMPTY(&(dfs->dfs_radarq))) {
		event = STAILQ_FIRST(&(dfs->dfs_radarq));
		STAILQ_REMOVE_HEAD(&(dfs->dfs_radarq), re_list);
		memset(event, 0, sizeof(struct dfs_event));
		STAILQ_INSERT_TAIL(&(dfs->dfs_eventq), event, re_list);
	}
	ATH_DFSEVENTQ_UNLOCK(dfs);
	ATH_DFSQ_UNLOCK(dfs);
}


/*
 * Add the radar event to the appropriate linked list which represents
 * the pulses which are currently in the tapped delay line with time
 * delay from the previous event
 */

static void
dfs_add_pulse(struct ath_softc *sc, struct dfs_filter *rf, struct dfs_event *re,
	      u_int32_t deltaT)
{
	u_int32_t index, n, window, pri;
	struct dfs_delayline *dl;	
	
	dl = &rf->rf_dl;
	/* Circular buffer of size 2^n */
	index = (dl->dl_lastelem + 1) & DFS_MAX_DL_MASK;
	if (dl->dl_numelems == DFS_MAX_DL_SIZE)
		dl->dl_firstelem = (dl->dl_firstelem + 1) & DFS_MAX_DL_MASK;
	else
		dl->dl_numelems++;
	dl->dl_lastelem = index;
	dl->dl_elems[index].de_time = deltaT;
	window = deltaT;
	dl->dl_elems[index].de_dur = re->re_dur;
	dl->dl_elems[index].de_rssi = re->re_rssi;

	for (n=0;n<dl->dl_numelems-1; n++) {
		index = (index-1) & DFS_MAX_DL_MASK;
		pri = dl->dl_elems[index].de_time;
		window += pri;
		if (window > rf->rf_filterlen) {
			dl->dl_firstelem = (index+1) & DFS_MAX_DL_MASK;
			dl->dl_numelems = n+1;
		}
	}
	DPRINTF(sc, ATH_DEBUG_DFS2,
		"dl firstElem = %d  lastElem = %d pri = %d dur = %d \n",
		dl->dl_firstelem, dl->dl_lastelem, deltaT, re->re_dur);
}
#if 0
static	int dfs_bin_fixedpattern_check(struct ath_softc *sc, struct dfs_filter *rf)
{
	struct ath_dfs *dfs=sc->sc_dfs;
	struct dfs_pulseline *pl = &dfs->pulses;
	int i, n, refpri, primargin, numpulses=1;
	u_int64_t this_ts, event_ts, window_start, window_end;
	u_int32_t index, event_dur;

	primargin = (rf->rf_maxpri - rf->rf_minpri)/2;
	refpri = (rf->rf_minpri + rf->rf_maxpri)/2;
	index = pl->pl_lastelem;
	this_ts = pl->pl_elems[index].p_time;
	for (n=1;n<rf->rf_numpulses; n++) {
		window_start = (this_ts - (refpri*n))-(primargin+n);
		window_end = window_start + (2*primargin)+n;
		DPRINTF(sc, ATH_DEBUG_DFS2,
		"window_start %u window_end %u \n",
		(u_int32_t)window_start, (u_int32_t)window_end);
		for(i=0;i<pl->pl_numelems;i++) {
			index = (index-1) & DFS_MAX_PULSE_BUFFER_MASK;
			event_ts = pl->pl_elems[index].p_time;
			DPRINTF(sc, ATH_DEBUG_DFS2,
			"ts %u \n", (u_int32_t)event_ts);
			event_dur = pl->pl_elems[index].p_dur;
			if( (event_ts <= window_end) && (event_ts >= window_start)){
				if( (event_dur >= rf->rf_mindur) && (event_dur <= rf->rf_maxdur)){
					numpulses++;	
					DPRINTF(sc, ATH_DEBUG_DFS2,
					"numpulses %u \n", numpulses);
					break;
				}
			}
			else if( event_ts < window_start){
				index = (index+1) & DFS_MAX_PULSE_BUFFER_MASK;
				break;
			}
		}
	}
	if( numpulses >= rf->rf_threshold)
		return 1;
	else
		return 0;
}
#endif

static	int dfs_bin_check(struct ath_softc *sc, struct dfs_filter *rf,
			     u_int32_t deltaT, u_int32_t width)
{
	u_int32_t refpri, refdur, searchpri, searchdur, deltapri, deltadur, averagerefpri;
	u_int32_t n, i, j, primargin, durmargin, highscore = 0, highscoreindex = 0;
	int found=0, numpulses=1;  //first pulse in the burst is most likely 
				   //being filtered out based on maxfilterlen
	int score[DFS_MAX_DL_SIZE];
	int rssi[DFS_MAX_DL_SIZE];
	int primultiples[6];
	int delayindex, dindex, rssimargin, averagerssi, deltarssi, totalrssi = 0;
	struct dfs_delayline *dl;
	u_int32_t scoreindex, lowpriindex= 0, lowpri = 0xffff;	
	dl = &rf->rf_dl;	
	if( dl->dl_numelems < (rf->rf_threshold-1))
		return 0;
	if( deltaT > rf->rf_filterlen)
		return 0;

	primargin = 6;
	if(rf->rf_maxdur < 10) {
		durmargin = 4;
	}
	else {
		durmargin = 6;
	}
	
	/* find out the average rssi */	
	rssimargin = 9;
	memset(rssi, 0, sizeof(int)*DFS_MAX_DL_SIZE);
	for (n=0;n<dl->dl_numelems; n++) {
		delayindex = (dl->dl_firstelem + n) & DFS_MAX_DL_MASK;
		totalrssi += dl->dl_elems[delayindex].de_rssi;
	}
	averagerssi = totalrssi/dl->dl_numelems;

	if( rf->rf_pritype == 1 ){
		refpri = (rf->rf_maxpri + rf->rf_minpri)/2;
		refdur = (rf->rf_maxdur + rf->rf_mindur)/2;
		goto refprifound;
#if 0
		found = dfs_bin_fixedpattern_check(sc, rf);
		if(found) {
			dl->dl_numelems = 0;
			DPRINTF(sc, ATH_DEBUG_DFS3, "matched filterID %d\n",rf->rf_pulseid);
		}
		return found;
#endif
	}
	memset(score, 0, sizeof(int)*DFS_MAX_DL_SIZE);

	/* find out the lowest pri */	
	for (n=0;n<dl->dl_numelems; n++) {
		delayindex = (dl->dl_firstelem + n) & DFS_MAX_DL_MASK;
		refpri = dl->dl_elems[delayindex].de_time;
		if(refpri < lowpri) {
			lowpri = dl->dl_elems[delayindex].de_time;
			lowpriindex = n;
		}
	}
	/* find out the each delay element's pri score */
	for (n=0;n<dl->dl_numelems; n++) {
		delayindex = (dl->dl_firstelem + n) & DFS_MAX_DL_MASK;
		refpri = dl->dl_elems[delayindex].de_time;
		for (i=0;i<dl->dl_numelems; i++) {
			dindex = (dl->dl_firstelem + i) & DFS_MAX_DL_MASK;
			searchpri = dl->dl_elems[dindex].de_time;
			deltapri = DFS_DIFF(searchpri, refpri);
			if( deltapri < primargin)
				score[n]++;
		}
		if( score[n] > rf->rf_threshold) {
			/* we got the most possible candidate,
			 * no need to continue further */
			break;
		}
	}
	/* find out the high scorer */
	for (n=0;n<dl->dl_numelems; n++) {
		if( score[n] > highscore) {
			highscore = score[n];
			highscoreindex = n;
		}
		else if( score[n] == highscore ) {
			/*more than one pri has highscore take the least pri */	
			delayindex = (dl->dl_firstelem + highscoreindex) & DFS_MAX_DL_MASK;
			dindex = (dl->dl_firstelem + n) & DFS_MAX_DL_MASK;
			if( dl->dl_elems[dindex].de_time <= 
				dl->dl_elems[delayindex].de_time ) {
				highscoreindex = n;		
			}
		}
	}
	/* find the average pri of pulses around the pri of highscore or
	 * the pulses around the lowest pri */
	if( highscore < 3) {	
		scoreindex = lowpriindex;
	}
	else { 
		scoreindex = highscoreindex;
	}

	/* We got the possible pri, save its parameters as reference */
	delayindex = (dl->dl_firstelem + scoreindex) & DFS_MAX_DL_MASK;
	refdur = dl->dl_elems[delayindex].de_dur;
	refpri = dl->dl_elems[delayindex].de_time;
	
	averagerefpri = 0;
	if( score[scoreindex] > 1) {
		for (i=0;i<dl->dl_numelems; i++) {
			dindex = (dl->dl_firstelem + i) & DFS_MAX_DL_MASK;
			searchpri = dl->dl_elems[dindex].de_time;
			deltapri = DFS_DIFF(searchpri, refpri);
			if( deltapri < primargin)
				averagerefpri += searchpri;
		}
		refpri = (averagerefpri/(score[scoreindex]));  //average

	}
refprifound:
	/* Note: Following primultiple calculation should be done once per filter 
	 * during initialization stage (dfs_attach) and stored in its array
	 * atleast for fixed frequency types like FCC Bin1 to save some CPU cycles.
	 * multiplication, devide operators in the following code are left as it is
	 * for readability hoping the complier will use left/right shifts wherever possible
	 */
	if( refpri > rf->rf_maxpri) {
		primultiples[0] = (refpri - refdur)/2;
		primultiples[1] = refpri;
		primultiples[2] = refpri + primultiples[0];
		primultiples[3] = (refpri - refdur)*2;
		primultiples[4] = primultiples[3] + primultiples[0];
		primultiples[5] = primultiples[3] + refpri;
	}
	else {
		primultiples[0] = refpri;
		primultiples[1] = refpri + primultiples[0];
		primultiples[2] = refpri + primultiples[1];
		primultiples[3] = refpri + primultiples[2];
		primultiples[4] = refpri + primultiples[3];
		primultiples[5] = refpri + primultiples[4];
	}
	DPRINTF(sc, ATH_DEBUG_DFS2,
		"pri0 = %d pri1 = %d pri2 = %d pri3 = %d pri4 = %d pri5 = %d\n",
		primultiples[0], primultiples[1], primultiples[2], 
		primultiples[3], primultiples[4], primultiples[5]);
	DPRINTF(sc, ATH_DEBUG_DFS2,
		"refpri = %d high score = %d index = %d numpulses = %d\n",
		refpri, highscore, highscoreindex, numpulses);

	/* Count the other delay elements that have  pri and dur with in the
	 * acceptable range from the reference one */
	for (i=0; i<dl->dl_numelems; i++) {
		delayindex = (dl->dl_firstelem + i) & DFS_MAX_DL_MASK;
		searchpri = dl->dl_elems[delayindex].de_time;
		searchdur = dl->dl_elems[delayindex].de_dur;
		deltadur = DFS_DIFF(searchdur, refdur);
		DPRINTF(sc, ATH_DEBUG_DFS2,
			"searchpri=%d index=%d searchdur=%d refdur=%d durmargin=%d\n",
			searchpri, i, searchdur, refdur, durmargin);
		for(j=0; j<6; j++) {
			deltapri = DFS_DIFF(searchpri, primultiples[j]);
			if( deltapri < (primargin+j)) {
				if( deltadur < durmargin) {
					deltarssi = DFS_DIFF(averagerssi, dl->dl_elems[delayindex].de_rssi);
					if( deltarssi < 10 ) {
						numpulses++;
						DPRINTF(sc, ATH_DEBUG_DFS2, "numpulses=%d\n", numpulses);
					}
				}
				break;
			}
		}
	}
	if( numpulses >= rf->rf_threshold){
		dl->dl_numelems = 0;
		found = 1;
		DPRINTF(sc, ATH_DEBUG_DFS3, "matched filterID %d\n",rf->rf_pulseid);
	}
	return found;
}

static void
dfs_nol_addchan(struct ath_softc *sc, HAL_CHANNEL *chan)
{
	struct ath_dfs *dfs=sc->sc_dfs;
	struct dfs_nolelem *nol, *elem, **next;
	u_int64_t tsf;

	if (dfs == NULL) {
		DPRINTF(sc, ATH_DEBUG_DFS, "%s: sc_dfs is NULL\n",
			__func__);
		return;
	}
	nol = dfs->dfs_nol;
	next = &(dfs->dfs_nol);
	tsf = ath_hal_gettsf64(sc->sc_ah);
	tsf += DFS_NOL_TIME;
	while (nol != NULL) {
		if ((nol->nol_chan.channel == chan->channel) &&
		    (nol->nol_chan.channelFlags == chan->channelFlags) &&
		    (nol->nol_chan.privFlags == chan->privFlags)) {
			nol->nol_tsffree = tsf;
			printk ("%s: Adding channel to NOL\n", __func__);
			ath_hal_dfsfound(sc->sc_ah, chan, DFS_NOL_TIME);
			return;
		}
		next = &(nol->nol_next);
		nol = nol->nol_next;
	}
	elem = kmalloc(sizeof(struct dfs_nolelem),GFP_KERNEL);
	if (elem == NULL) {
		DPRINTF(sc, ATH_DEBUG_DFS,
			"%s: failed to allocate memory for nol entry\n",
			 __func__);
		return;
	}
	memset(elem, 0, sizeof(struct dfs_nolelem));
	elem->nol_chan = *chan;
	elem->nol_tsffree = tsf;
	elem->nol_next = NULL;
	*next = elem;
}

static int
dfs_bin5_addpulse(struct ath_softc *sc, struct dfs_bin5radars *br,
		  struct dfs_event *re, u_int64_t thists)
{
	u_int32_t index = 0,stop;
	u_int64_t tsDelta = 1;

	/* Check if this pulse is a valid pulse in terms of repetition, 
	 * if not, return without adding it to the queue.
	 * PRI : Pulse Repetition Interval
	 * BRI : Burst Repetition Interval */ 
	if( br->br_numelems != 0){
		index = br->br_lastelem;
		tsDelta = thists - br->br_elems[index].be_ts;
		if( (tsDelta < DFS_BIN5_PRI_LOWER_LIMIT && tsDelta != 0) ||
			( (tsDelta > DFS_BIN5_PRI_HIGHER_LIMIT) &&
			  (tsDelta < DFS_BIN5_BRI_LOWER_LIMIT))) {
			return 0;
		}
	}
	/* Circular buffer of size 2^n */
	if( tsDelta == 0 ) {
		/* overwriting the previous pulse.
		   No need to increment count */
	}
	else {
		index = (br->br_lastelem +1) & DFS_MAX_B5_MASK;
		br->br_lastelem = index;
		if (br->br_numelems == DFS_MAX_B5_SIZE)
			br->br_firstelem = (br->br_firstelem+1)&DFS_MAX_B5_MASK;
		else
			br->br_numelems++;
	}
	br->br_elems[index].be_ts = thists;
	br->br_elems[index].be_rssi = re->re_rssi;
	br->br_elems[index].be_dur = re->re_dur;
	
	stop = 0;
	index = br->br_firstelem;
	while ((!stop) && (br->br_numelems-1) > 0) {
		if ((thists - br->br_elems[index].be_ts) > 
		    ((u_int64_t) br->br_pulse.b5_timewindow)) {
			br->br_numelems--;
			br->br_firstelem = (br->br_firstelem +1) & DFS_MAX_B5_MASK;
			index = br->br_firstelem;
		} else
			stop = 1;
	}
	if( br->br_numelems >= (br->br_pulse.b5_threshold*2)) {
		return 1;
	}
	else {
		/* Added this pulse to the delay line succesfullly, but
		 * there are not enough pulses to meet the threshold.So, no need to 
		 * check for pattern matching, return 0 */
		return 0;
	}
}

/*
 * If the dfs structure is NULL (which should be illegal if everyting is working
 * properly, then signify that a bin5 radar was found
 */

static int
dfs_bin5_check(struct ath_softc *sc)
{
	struct ath_dfs *dfs=sc->sc_dfs;
	struct dfs_bin5radars *br;
	int index[DFS_MAX_B5_SIZE];
	u_int32_t numelems, n, i, this, prev, rssi_diff, width_diff, bursts= 0;
	u_int32_t total_diff=0, average_diff, total_width=0, average_width, numevents=0;
	u_int32_t width_margin;
	u_int64_t pri;

	if (dfs == NULL) {
		DPRINTF(sc, ATH_DEBUG_DFS,"%s: sc_dfs is NULL\n",
			__func__);
		return 1;
	}

	if( IS_5413(sc->sc_ah)) {
		width_margin = DFS_BIN5_WIDTH_MARGIN_5413;
	}
	else {
		width_margin = DFS_BIN5_WIDTH_MARGIN;
	}
	for (n=0;n<dfs->dfs_rinfo.rn_numbin5radars; n++) {
		br = &(dfs->dfs_b5radars[n]);
		DPRINTF(sc, ATH_DEBUG_DFS2,
			"Num elems = %d\n",br->br_numelems);
		prev = br->br_firstelem;
		numelems = br->br_numelems;

		this = br->br_firstelem;
		DPRINTF(sc, ATH_DEBUG_DFS2,
			"First elem: ts%u %u %u\n",
			(u_int32_t)br->br_elems[this].be_ts,
			br->br_elems[this].be_dur,
			br->br_elems[this].be_rssi);

		for(i=1;i<numelems;i++){
			this = ((br->br_firstelem +i) & DFS_MAX_B5_MASK);
			/* Rule 1: 1000 <= PRI <= 2000 + some margin */
			if( br->br_elems[this].be_ts >= br->br_elems[prev].be_ts ) {
				pri = br->br_elems[this].be_ts - br->br_elems[prev].be_ts;
			}
			else {//roll over case
				//pri = (0xffffffffffffffff - br->br_elems[prev].be_ts) + br->br_elems[this].be_ts;
				pri = br->br_elems[this].be_ts;
			}
			DPRINTF(sc, ATH_DEBUG_DFS2, "\nthis_ts: %u pri: %u", 
					(u_int32_t)br->br_elems[this].be_ts, (u_int32_t)pri);
			if( (pri >= DFS_BIN5_PRI_LOWER_LIMIT) && (pri <= DFS_BIN5_PRI_HIGHER_LIMIT)) {
				/* Rule 2: pulse width of the pulses in the burst should be same (+/- margin) */
				width_diff = DFS_DIFF(br->br_elems[this].be_dur, br->br_elems[prev].be_dur);
				DPRINTF(sc, ATH_DEBUG_DFS2, " width_diff:%u", width_diff);
				if( width_diff <= width_margin ) {
					/* Rule 3: RSSI of the pulses in the burst should be same (+/- margin) */
					rssi_diff = DFS_DIFF(br->br_elems[this].be_rssi, br->br_elems[prev].be_rssi);
					if( rssi_diff <= DFS_BIN5_RSSI_MARGIN ) {
						bursts++;
						DPRINTF(sc, ATH_DEBUG_DFS2, " num_bursts:%d\n", bursts);
						/* Save the indexes of this pair for later width variance check */
						if( numevents >= 2 ) {
							/* make sure the event is not duplicated,
							 * possible in a 3 pulse burst */
							if( index[numevents-1] != prev) {
								index[numevents++] = prev;
							}
						}
						else {
							index[numevents++] = prev;
						}
						index[numevents++] = this;
					}
				}
			}
			prev = this;
		}
		if ( bursts >= br->br_pulse.b5_threshold) {
			if( (br->br_elems[br->br_lastelem].be_ts - br->br_elems[br->br_firstelem].be_ts) < 3000000 ) {
				return 0;
			}
			else {
				for (i=0; i<numevents; i++){
					total_width += br->br_elems[index[i]].be_dur;
				}
				average_width = total_width/numevents;
				for (i=0; i<numevents; i++){
					total_diff += DFS_DIFF(br->br_elems[index[i]].be_dur, average_width);
				}
				average_diff = total_diff/numevents;
				if( average_diff > 4 ) {
					DPRINTF(sc, ATH_DEBUG_DFS3, "%s BIN5 radar found\n", __func__);
					return 1;
				}
			}
		}
	}
	return 0;
}

static inline u_int8_t
dfs_process_pulse_dur(struct ath_softc *sc, u_int8_t re_dur) 
{
        struct ath_dfs *dfs=sc->sc_dfs;

	if (re_dur == 0 || 
              (dfs->dfsdomain == DFS_ETSI_DOMAIN &&
                re_dur <= 6 /* MKK, ETSI hack */)) {
		return 1;
	} else {
		/* Convert 0.8us durations to TSF ticks (usecs) */
		return (u_int8_t)_round((int32_t)(80*re_dur));
	}
}

struct dfs_state *
dfs_getchanstate(struct ath_softc *sc, u_int8_t *index)
{
	struct ath_dfs *dfs=sc->sc_dfs;
	struct dfs_state *rs=NULL;
	int i;

	if (dfs == NULL) {
		DPRINTF(sc, ATH_DEBUG_DFS,"%s: sc_dfs is NULL\n",
			__func__);
		return NULL;
	}
	for (i=0;i<DFS_NUM_RADAR_STATES; i++) {
		if ((dfs->dfs_radar[i].rs_chan.channel == sc->sc_curchan.channel) &&
		    (dfs->dfs_radar[i].rs_chan.channelFlags == sc->sc_curchan.channelFlags)) {
			if (index != NULL)
				*index = (u_int8_t) i;
			return &(dfs->dfs_radar[i]);
		}
	}
	/* No existing channel found, look for first free channel state entry */
	for (i=0; i<DFS_NUM_RADAR_STATES; i++) {
		if (dfs->dfs_radar[i].rs_chan.channel == 0) {
			rs = &(dfs->dfs_radar[i]);
			/* Found one, set channel info and default thresholds */
			rs->rs_chan = sc->sc_curchan;
			rs->rs_firpwr = dfs->dfs_defaultparams.pe_firpwr;
			rs->rs_radarrssi = dfs->dfs_defaultparams.pe_rrssi;
			rs->rs_height = dfs->dfs_defaultparams.pe_height;
			rs->rs_pulserssi = dfs->dfs_defaultparams.pe_prssi;
			rs->rs_inband = dfs->dfs_defaultparams.pe_inband;
			/* 5413 specific */
			rs->rs_relpwr = dfs->dfs_defaultparams.pe_relpwr;
			rs->rs_relstep = dfs->dfs_defaultparams.pe_relstep;
			rs->rs_maxlen = dfs->dfs_defaultparams.pe_maxlen;

			if (index != NULL)
				*index = (u_int8_t) i;
			return (rs);
		}
	}
	DPRINTF(sc, ATH_DEBUG_DFS, "%s: No more radar states left.\n", __func__);
	return(NULL);
}
EXPORT_SYMBOL(dfs_getchanstate);

static void dfs_dump_pulselog( struct dfs_pulseline *pl)
{
	u_int32_t i, index;
	/* dump the pulse log on to console */
	for(i=0; i < pl->pl_numelems; i++ ) {
		index = (pl->pl_firstelem + i) & DFS_MAX_PULSE_BUFFER_MASK;
		//printk("ts %u dur %u rssi %u\n",
		printk("ts%u %u %u\n",
		(u_int32_t)pl->pl_elems[index].p_time,
		pl->pl_elems[index].p_dur,
		pl->pl_elems[index].p_rssi);
	}
}
int
dfs_process_radarevent(struct ath_softc *sc, HAL_CHANNEL *chan)
{
#ifdef ATH_SUPPORT_DFS
	struct ath_dfs *dfs=sc->sc_dfs;
	struct ath_hal *ah=sc->sc_ah;
	struct dfs_event re,*event;
	struct dfs_state *rs=NULL;
	struct dfs_filtertype *ft;
	struct dfs_filter *rf;
	int found, retval=0,p, empty;
	u_int32_t tabledepth,rfilt, index;
	u_int64_t deltafull_ts,this_ts, deltaT;
	HAL_CHANNEL *thischan;
	HAL_PHYERR_PARAM pe;
	struct dfs_pulseline *pl = &dfs->pulses;

	if ( ! (sc->sc_curchan.channelFlags & CHANNEL_DFS) ) 
		return 0;

	if (dfs == NULL) {
		DPRINTF(sc, ATH_DEBUG_DFS, "%s: sc_sfs is NULL\n",
			__func__);
		return 1;
	}

 	/* TEST : Simulate radar bang */
 	if (sc->sc_ic.ic_bangradar) {
 		*chan = sc->sc_curchan;
 		sc->sc_ic.ic_bangradar = 0; //reset
                dfs->stats.num_radar_detects++;
		if (sc->sc_debug & ATH_DEBUG_DFS3) {
			/* Disable radar detection until we dump the log */
			rfilt = ath_hal_getrxfilter(ah);
			rfilt &= ~HAL_RX_FILTER_PHYRADAR;
			ath_hal_setrxfilter(ah, rfilt);
			dfs_dump_pulselog(pl);
			/* enable radar detection */
			rfilt = ath_hal_getrxfilter(ah);
			rfilt |= HAL_RX_FILTER_PHYRADAR;
			ath_hal_setrxfilter(ah, rfilt);
		}
 		return 1;
 	}

	ATH_DFSQ_LOCK(dfs);
	empty = STAILQ_EMPTY(&(dfs->dfs_radarq));
	ATH_DFSQ_UNLOCK(dfs);

	while ((!empty) && (!retval)) {
		ATH_DFSQ_LOCK(dfs);
		event = STAILQ_FIRST(&(dfs->dfs_radarq));
		if (event != NULL)
			STAILQ_REMOVE_HEAD(&(dfs->dfs_radarq), re_list);
		ATH_DFSQ_UNLOCK(dfs);

		if (event == NULL) {
			empty = 1;
			break;
		}
		re = *event;

		memset(event, 0, sizeof(struct dfs_event));
		ATH_DFSEVENTQ_LOCK(dfs);
		STAILQ_INSERT_TAIL(&(dfs->dfs_eventq), event, re_list);
		ATH_DFSEVENTQ_UNLOCK(dfs);

		found = 0;
		if (re.re_chanindex < DFS_NUM_RADAR_STATES)
			rs = &dfs->dfs_radar[re.re_chanindex];
		else {
			ATH_DFSQ_LOCK(dfs);
			empty = STAILQ_EMPTY(&(dfs->dfs_radarq));
			ATH_DFSQ_UNLOCK(dfs);
			continue;
		}
		if (rs->rs_chan.privFlags & CHANNEL_INTERFERENCE) {
			ATH_DFSQ_LOCK(dfs);
			empty = STAILQ_EMPTY(&(dfs->dfs_radarq));
			ATH_DFSQ_UNLOCK(dfs);
			continue;
		}

		if (dfs->dfs_rinfo.rn_lastfull_ts == 0) {
			/*
			 * Either not started, or 64-bit rollover exactly to zero
			 * Just prepend zeros to the 15-bit ts
			 */
			dfs->dfs_rinfo.rn_ts_prefix = 0;
			this_ts = (u_int64_t) re.re_ts;
		} else {
			if (re.re_ts <= dfs->dfs_rinfo.rn_last_ts) {
				dfs->dfs_rinfo.rn_ts_prefix += 
					(((u_int64_t) 1) << DFS_TSSHIFT);
				/* Now, see if it's been more than 1 wrap */
				deltafull_ts = re.re_full_ts - dfs->dfs_rinfo.rn_lastfull_ts;
				if (deltafull_ts > 
				    ((u_int64_t)((DFS_TSMASK - dfs->dfs_rinfo.rn_last_ts) + 1 + re.re_ts)))
					deltafull_ts -= (DFS_TSMASK - dfs->dfs_rinfo.rn_last_ts) + 1 + re.re_ts;
				deltafull_ts = deltafull_ts >> DFS_TSSHIFT;
				if (deltafull_ts > 1) {
					dfs->dfs_rinfo.rn_ts_prefix += 
						((deltafull_ts - 1) << DFS_TSSHIFT);
				}
			} else {
				deltafull_ts = re.re_full_ts - dfs->dfs_rinfo.rn_lastfull_ts;
				if (deltafull_ts > (u_int64_t) DFS_TSMASK) {
					deltafull_ts = deltafull_ts >> DFS_TSSHIFT;
					dfs->dfs_rinfo.rn_ts_prefix += 
						((deltafull_ts - 1) << DFS_TSSHIFT);
				}
			}
			this_ts = dfs->dfs_rinfo.rn_ts_prefix | ((u_int64_t) re.re_ts);
		}
		dfs->dfs_rinfo.rn_lastfull_ts = re.re_full_ts;
		dfs->dfs_rinfo.rn_last_ts = re.re_ts;

		re.re_dur = dfs_process_pulse_dur(sc, re.re_dur);
		
		DPRINTF(sc, (ATH_DEBUG_DFS1 | ATH_DEBUG_DFS3),"ts%u %u %u\n",
			(u_int32_t)this_ts, re.re_dur, re.re_rssi);


		/* Save the pulse parameters in the pulse buffer(pulse line) */
		index = pl->pl_lastelem;
		deltaT = this_ts - (pl->pl_elems[index].p_time + pl->pl_elems[index].p_dur);
		if( (deltaT <= 10 ) && (IS_5413(ah))) {
			this_ts =  pl->pl_elems[index].p_time;
			re.re_dur += (pl->pl_elems[index].p_dur+(u_int8_t)deltaT);
			re.re_rssi = (pl->pl_elems[index].p_rssi+re.re_rssi)/2;
			DPRINTF(sc, ATH_DEBUG_DFS1,"combined pulse : ts%u %u %u\n",
				(u_int32_t)this_ts, re.re_dur, re.re_rssi);
		}
		else {
			index = (pl->pl_lastelem + 1) & DFS_MAX_PULSE_BUFFER_MASK;
			if (pl->pl_numelems == DFS_MAX_PULSE_BUFFER_SIZE)
				pl->pl_firstelem = (pl->pl_firstelem+1) & DFS_MAX_PULSE_BUFFER_MASK;
			else
				pl->pl_numelems++;
			pl->pl_lastelem = index;
		}
		pl->pl_elems[index].p_time = this_ts;
		pl->pl_elems[index].p_dur = re.re_dur;
		pl->pl_elems[index].p_rssi = re.re_rssi;

		found = 0;
		for (p=0; (p<dfs->dfs_rinfo.rn_numbin5radars)&&(!found); p++) {
			struct dfs_bin5radars *br;
			u_int32_t b5_rssithresh;
	
			br = &(dfs->dfs_b5radars[p]);
			b5_rssithresh = br->br_pulse.b5_rssithresh;

                        /* Adjust the filter threshold for rssi in non TURBO mode*/
                        if( ! (sc->sc_curchan.channelFlags & CHANNEL_TURBO) ) {
                                b5_rssithresh += br->br_pulse.b5_rssimargin;
                        }

			if ((re.re_dur >= br->br_pulse.b5_mindur) &&
			    (re.re_dur <= br->br_pulse.b5_maxdur) &&
			    (re.re_rssi >= b5_rssithresh)) {
				if( dfs_bin5_addpulse(sc, br, &re, this_ts) ) {
					found |= dfs_bin5_check(sc);
				}
			}
		}
		if (found) {
			DPRINTF(sc, ATH_DEBUG_DFS1, "%s: Found bin5 radar\n", __func__);
			retval |= found;
			goto dfsfound;
		}
		tabledepth = 0;
		rf = NULL;
		DPRINTF(sc, ATH_DEBUG_DFS2,"  ** RD (%d): ts %u dur %u rssi %u\n",
		       rs->rs_chan.channel, re.re_ts, re.re_dur, re.re_rssi);
		while ((tabledepth < DFS_MAX_RADAR_OVERLAP) &&
		       ((dfs->dfs_radartable[re.re_dur])[tabledepth] != -1) &&
		       (!retval)) {
			ft = dfs->dfs_radarf[((dfs->dfs_radartable[re.re_dur])[tabledepth])];
			
			if (re.re_rssi < ft->ft_rssithresh ) {
				DPRINTF(sc, ATH_DEBUG_DFS2, "%s: Rejecting on rssi\n", __func__);
				tabledepth++;
				ATH_DFSQ_LOCK(dfs);
				empty = STAILQ_EMPTY(&(dfs->dfs_radarq));
				ATH_DFSQ_UNLOCK(dfs);
				continue;
			}
			deltaT = this_ts - ft->ft_last_ts;
			DPRINTF(sc, ATH_DEBUG_DFS2, "deltaT = %lld (ts: %llu)  (last ts: %llu)\n",
				deltaT, this_ts, ft->ft_last_ts);
			if (deltaT < ft->ft_minpri) {
				/* This check is for the whole filter type. Individual filters 
				 * will check this again later. This is first line of filtering */
				if( deltaT < 10 ) {
					ft->ft_last_ts = this_ts;
					DPRINTF(sc, ATH_DEBUG_DFS2,
					"%s: Rejecting on filtertype PRI < 10us\n", __func__);
				}
				else {
					DPRINTF(sc, ATH_DEBUG_DFS2, 
					"Reject on filtertype min PRI %u\n", ft->ft_minpri);
				}
				tabledepth++;
				continue;
			}
			for (p=0, found = 0; (p<ft->ft_numfilters) && (!found); p++) {
				rf = &(ft->ft_filters[p]);
				if ((re.re_dur >= rf->rf_mindur) && (re.re_dur <= rf->rf_maxdur)) {
					/* The above check is probably not necessary */
					deltaT = this_ts - rf->rf_dl.dl_last_ts;
					if (deltaT < 0)
						deltaT = (int64_t) ((DFS_TSF_WRAP - rf->rf_dl.dl_last_ts) + this_ts +1);
					
					if (deltaT < rf->rf_minpri) {
						/* Second line of PRI filtering. */
						DPRINTF(sc, ATH_DEBUG_DFS2, 
						"filterID %d : Rejecting on individual filter min PRI\n",
						rf->rf_pulseid);
						continue;
					}

					if ((deltaT > rf->rf_maxpri) && (deltaT < (2*rf->rf_minpri))) {
						DPRINTF(sc, ATH_DEBUG_DFS2, 
						"filterID %d maxpri %d : Rejecting on max PRI: %lld\n",
						rf->rf_pulseid, rf->rf_maxpri, deltaT);
						/* But update the last time stamp */
						rf->rf_dl.dl_last_ts = this_ts;
						continue;
					}
					dfs_add_pulse(sc, rf, &re, deltaT);
					found = dfs_bin_check(sc, rf, (u_int32_t) deltaT, re.re_dur);
					if ((sc->sc_debug) & ATH_DEBUG_DFS2) {
						dfs_print_delayline(sc, &rf->rf_dl);
					}
					rf->rf_dl.dl_last_ts = this_ts;
				}
			}

			ft->ft_last_ts = this_ts;
			retval |= found;
			if (found) {
				DPRINTF(sc, (ATH_DEBUG_DFS1|ATH_DEBUG_DFS3),
					"Found Radar at ts %llu\n",this_ts);
			}
			tabledepth++;
		}
		ATH_DFSQ_LOCK(dfs);
		empty = STAILQ_EMPTY(&(dfs->dfs_radarq));
		ATH_DFSQ_UNLOCK(dfs);
	}
dfsfound:
	if (retval) {
		thischan = &rs->rs_chan;
		if ((dfs->dfs_rinfo.rn_use_nol) &&
		    (sc->sc_ic.ic_opmode == IEEE80211_M_HOSTAP))
			thischan->privFlags |= CHANNEL_INTERFERENCE;
		chan->channel = thischan->channel;
		chan->channelFlags = thischan->channelFlags;
		chan->privFlags  = thischan->privFlags;
		chan->maxRegTxPower = thischan->maxRegTxPower;
		if ((dfs->dfs_rinfo.rn_use_nol) &&
		    (sc->sc_ic.ic_opmode == IEEE80211_M_HOSTAP))
			dfs_nol_addchan(sc, thischan);

		/* Disable radar for now */
		rfilt = ath_hal_getrxfilter(ah);
		rfilt &= ~HAL_RX_FILTER_PHYRADAR;
		ath_hal_setrxfilter(ah, rfilt);

		if (sc->sc_debug & ATH_DEBUG_DFS3) {
			dfs_dump_pulselog(pl);
		}

		dfs_reset_radarq(sc);
		dfs_reset_alldelaylines(sc);
			
		/* XXX Should we really enable again? Maybe not... */
		pe.pe_firpwr = rs->rs_firpwr;
		pe.pe_rrssi = rs->rs_radarrssi;
		pe.pe_height = rs->rs_height;
		pe.pe_prssi = rs->rs_pulserssi;
		pe.pe_inband = rs->rs_inband;
		/* 5413 specific */
		pe.pe_relpwr = rs->rs_relpwr;
		pe.pe_relstep = rs->rs_relstep;
		pe.pe_maxlen = rs->rs_maxlen;

		ath_hal_enabledfs(ah, &pe);
		rfilt |= HAL_RX_FILTER_PHYRADAR;
		ath_hal_setrxfilter(ah, rfilt);
		if ((sc->sc_curchan.channel == chan->channel) &&
		    (sc->sc_curchan.channelFlags == chan->channel)) {
			if ((chan->privFlags & CHANNEL_INTERFERENCE) &&
			    (dfs->dfs_rinfo.rn_use_nol))
				sc->sc_curchan.privFlags |= CHANNEL_INTERFERENCE;
		}
                /* Collect stats */
                dfs->stats.num_radar_detects++;
	}
	return retval;
#endif /* ATH_SUPPORT_DFS */
	return 0;

}
EXPORT_SYMBOL(dfs_process_radarevent);

u_int32_t
dfs_check_nol(struct ath_softc *sc, HAL_CHANNEL *chans, u_int32_t nchans)
{
#ifdef ATH_SUPPORT_DFS
	struct ath_dfs *dfs=sc->sc_dfs;
	struct dfs_nolelem *nol,**prev_next;
	u_int64_t tsf;
	u_int32_t free_index=0;

	if (dfs == NULL) {
		DPRINTF(sc, ATH_DEBUG_DFS, "%s: sc_dfs is NULL\n",
			__func__);
		return 0;
	}
	tsf = ath_hal_gettsf64(sc->sc_ah);
	prev_next = &(dfs->dfs_nol);
	nol = dfs->dfs_nol;
	while (nol != NULL) {
		if (tsf >= nol->nol_tsffree) {
			*prev_next = nol->nol_next;
			nol->nol_chan.privFlags &= ~CHANNEL_INTERFERENCE;
			if (free_index < nchans) {
				chans[free_index].channel = nol->nol_chan.channel;
				chans[free_index].channelFlags = nol->nol_chan.channelFlags;
				chans[free_index].privFlags = nol->nol_chan.privFlags;
				free_index++;
			}
			ath_hal_checkdfs(sc->sc_ah, &nol->nol_chan);
			kfree(nol);
			nol = *prev_next;
		} else {
			prev_next = &(nol->nol_next);
			nol = nol->nol_next;
		}
	}
	return free_index;
#else
	return 0;
#endif /* AH_SUPPORT_DFS */
}
EXPORT_SYMBOL(dfs_check_nol);


static struct dfs_pulse *
get_dfs_radars(struct ath_softc *sc, u_int32_t dfsdomain, int *numradars,
	       struct dfs_bin5pulses **bin5pulses, int *numb5radars, HAL_PHYERR_PARAM *pe)
{
	struct dfs_pulse *dfs_radars=NULL;

#define	N(a)	(sizeof(a)/sizeof(a[0]))
	switch (dfsdomain) {
	case DFS_FCC_DOMAIN:
		if (IS_5413(sc->sc_ah)) {
			dfs_radars = &ar5413_fcc_radars[2];
			*numradars= N(ar5413_fcc_radars)-2;
			*bin5pulses = &ar5413_bin5pulses[0];
			*numb5radars = N(ar5413_bin5pulses);
		} else {
			dfs_radars = &ar5312_fcc_radars[2];
			*numradars = N(ar5312_fcc_radars)-2;
			*bin5pulses = &ar5312_bin5pulses[0];
			*numb5radars = N(ar5312_bin5pulses);
		}
		break;
	case DFS_ETSI_DOMAIN:
		if (IS_5413(sc->sc_ah)) {
			dfs_radars = &ar5413_etsi_radars[0];
			*numradars = N(ar5413_etsi_radars);
		} else {
			dfs_radars = &ar5312_etsi_radars[0];
			*numradars = N(ar5312_etsi_radars);
		}
		break;
	case DFS_MKK4_DOMAIN:
                if (IS_5413(sc->sc_ah)) {
                        dfs_radars = &ar5413_fcc_radars[0];
                        *numradars = N(ar5413_fcc_radars);
			*bin5pulses = &ar5413_bin5pulses[0];
			*numb5radars = N(ar5413_bin5pulses);
                } else {
                        dfs_radars = &ar5312_fcc_radars[0];
                        *numradars = N(ar5312_fcc_radars);
			*bin5pulses = &ar5312_bin5pulses[0];
			*numb5radars = N(ar5312_bin5pulses);
                }
		break;
	default:
		return NULL;
	}

	/* Set the default phy parameters per chip */
	if (IS_5413(sc->sc_ah)) {
		pe->pe_firpwr = AR5413_DFS_FIRPWR;
		pe->pe_rrssi = AR5413_DFS_RRSSI;
		pe->pe_height = AR5413_DFS_HEIGHT;
		pe->pe_prssi = AR5413_DFS_PRSSI;
		pe->pe_inband = AR5413_DFS_INBAND;
		pe->pe_relpwr = AR5413_DFS_RELPWR;
		pe->pe_relstep = AR5413_DFS_RELSTEP;
		pe->pe_maxlen = AR5413_DFS_MAXLEN;

	} else {
		/* Use the 5212/5312 DFS parameters */
		pe->pe_firpwr = AR5212_DFS_FIRPWR;
		pe->pe_rrssi = AR5212_DFS_RRSSI;
		pe->pe_height = AR5212_DFS_HEIGHT;
		pe->pe_prssi = AR5212_DFS_PRSSI;
		pe->pe_inband = AR5212_DFS_INBAND;
	}
	return dfs_radars;
#undef N
}
			
int
dfs_attach(struct ath_softc *sc)
{
	struct ath_dfs *dfs = sc->sc_dfs;
#ifdef ATH_SUPPORT_DFS
#define	N(a)	(sizeof(a)/sizeof(a[0]))
	int p,n;
	int numradars = 0, numb5radars = 0;
	u_int32_t T,i;
	struct dfs_filtertype *ft = NULL;
	struct dfs_filter *rf=NULL;
	struct dfs_pulse *dfs_radars;
	struct dfs_event *events = NULL;
	struct dfs_bin5pulses *b5pulses=NULL;
	int32_t min_rssithresh=DFS_MAX_RSSI_VALUE;
	u_int32_t max_pulsedur=0;
	
	/* to display the filter details at boot, uncomment the following line*/
	//sc->sc_debug |= (ATH_DEBUG_DFS2|ATH_DEBUG_DFS);

	if (dfs != NULL) {
		DPRINTF(sc, ATH_DEBUG_DFS, "%s: sc_dfs was not NULL\n",
			__func__);
		return 1;
	}
	dfs = kmalloc(sizeof(struct ath_dfs), GFP_KERNEL);
	if (dfs == NULL) {
		DPRINTF(sc, ATH_DEBUG_DFS,
			"%s: ath_dfs allocation failed\n", __func__);
		return 1;
	}
	memset(dfs, 0, sizeof (struct ath_dfs));
	sc->sc_dfs = dfs;
	ATH_DFSQ_LOCK_INIT(dfs);
	STAILQ_INIT(&dfs->dfs_radarq);
	ATH_ARQ_LOCK_INIT(dfs);
	STAILQ_INIT(&dfs->dfs_arq);
	STAILQ_INIT(&(dfs->dfs_eventq));
	ATH_DFSEVENTQ_LOCK_INIT(dfs);
	events = kmalloc(sizeof(struct dfs_event)*DFS_MAX_EVENTS,
			 GFP_KERNEL);
	if (events == NULL) {
		kfree(dfs);
		DPRINTF(sc, ATH_DEBUG_DFS,
			"%s: events allocation failed\n", __func__);
		return 1;
	}

	dfs->pulses.pl_lastelem = DFS_MAX_PULSE_BUFFER_MASK;
	for (i=0; i<DFS_MAX_EVENTS; i++) {
		STAILQ_INSERT_TAIL(&(dfs->dfs_eventq), &events[i], re_list);
	}
	/* XXX - KCYU - OK HERE */
	if (ath_hal_getcapability(sc->sc_ah, HAL_CAP_PHYDIAG,
				  HAL_CAP_AR, NULL) == HAL_OK) {
		dfs_reset_ar(sc);
		dfs_reset_arq(sc);
		dfs->dfs_proc_phyerr |= DFS_AR_EN;
	}
	/* OK here */
	if (ath_hal_getcapability(sc->sc_ah, HAL_CAP_PHYDIAG,
				  HAL_CAP_RADAR, NULL) == HAL_OK) {
		u_int32_t val;
		/* 
		 * If we have fast diversity capability, read off
		 * Strong Signal fast diversity count set in the ini
		 * file, and store so we can restore the value when
		 * radar is disabled
		 */
		if (ath_hal_getcapability(sc->sc_ah, HAL_CAP_DIVERSITY, HAL_CAP_STRONG_DIV,
					  &val) == HAL_OK) {
			dfs->dfs_rinfo.rn_fastdivGCval = val;
		}
		ath_hal_getcapability(sc->sc_ah, HAL_CAP_DFS_DMN, 0, &dfs->dfsdomain);
		if (domainoverride != DFS_UNINIT_DOMAIN) {
			printk (KERN_INFO "Overriding DFS domain with %d\n",domainoverride);
			dfs->dfsdomain = domainoverride;
		}
		dfs_radars = get_dfs_radars(sc, dfs->dfsdomain, &numradars, &b5pulses, &numb5radars,
					    &dfs->dfs_defaultparams);
		if (dfs_radars == NULL) {
			DPRINTF(sc, ATH_DEBUG_DFS, "%s: Unknownn dfs domain %d\n",
				 __func__, dfs->dfsdomain);
			/* Disable radar detection since we don't have a radar domain */
			dfs->dfs_proc_phyerr &= ~DFS_RADAR_EN;
			return 0;
		}
		sc->sc_ic.ic_isdfsregdomain = 1;
		dfs->dfs_proc_phyerr |= DFS_RADAR_EN;

		for (n=0; n<DFS_MAX_RADAR_TYPES; n++) {
			dfs->dfs_radarf[n] = kmalloc(sizeof(struct dfs_filtertype),GFP_KERNEL);
			if (dfs->dfs_radarf[n] == NULL) {
				DPRINTF(sc,ATH_DEBUG_DFS,
					"%s: cannot allocate memory for radar filter types\n",
					__func__);
				goto bad1;
			}
		}

		dfs->dfs_radartable = kmalloc(256*sizeof(int8_t *), GFP_KERNEL);
		if (dfs->dfs_radartable == NULL) {
			DPRINTF(sc, ATH_DEBUG_DFS, "%s: cannot allocate memory for radar table\n",
				__func__);
			goto bad1;
		}
		for (n=0; n<256; n++) {
			dfs->dfs_radartable[n] = kmalloc(DFS_MAX_RADAR_OVERLAP*sizeof(int8_t),
							 GFP_KERNEL);
			if (dfs->dfs_radartable[n] == NULL) {
				DPRINTF(sc, ATH_DEBUG_DFS,
					"%s: cannot allocate memory for radar table entry\n",
					__func__);
				goto bad2;
			}
			/* Clear filter type table */
			for (i=0;i<DFS_MAX_RADAR_OVERLAP; i++)
				(dfs->dfs_radartable[n])[i] = -1;
		}
		if (!usenol) {
			printk (KERN_INFO "\n\n\n\n\nDisabling Channel NOL\n\n\n\n\n");
		}
		dfs->dfs_rinfo.rn_use_nol = usenol;
		if ( dfs->dfs_rinfo.rn_use_nol )
			sc->sc_ic.ic_flags_ext |= IEEE80211_FEXT_MARKDFS;
		else
			sc->sc_ic.ic_flags_ext &= ~IEEE80211_FEXT_MARKDFS;
		dfs->dfs_rinfo.rn_numradars = 0;
		/* Now, initialize the radar filters */
		for (p=0; p<numradars; p++) {
			ft = NULL;
			for (n=0; n<dfs->dfs_rinfo.rn_numradars; n++) {
				if ((dfs_radars[p].rp_pritype == dfs->dfs_radarf[n]->ft_pritype) &&
				    (dfs_radars[p].rp_numpulses == dfs->dfs_radarf[n]->ft_numpulses) &&
				    (dfs_radars[p].rp_mindur == dfs->dfs_radarf[n]->ft_mindur) &&
				    (dfs_radars[p].rp_maxdur == dfs->dfs_radarf[n]->ft_maxdur)) {
					ft = dfs->dfs_radarf[n];
					break;
				}
			}
			if (ft == NULL) {

				/* No filter of the appropriate dur was found */
				if ((dfs->dfs_rinfo.rn_numradars+1) >DFS_MAX_RADAR_TYPES) {
					DPRINTF(sc, ATH_DEBUG_DFS, "%s: Too many filter types\n",
						__func__);
					goto bad4;
				}
				ft = dfs->dfs_radarf[dfs->dfs_rinfo.rn_numradars];
				ft->ft_numfilters = 0;
				ft->ft_numpulses = dfs_radars[p].rp_numpulses;
				ft->ft_mindur = dfs_radars[p].rp_mindur;
				ft->ft_maxdur = dfs_radars[p].rp_maxdur;
				ft->ft_pritype = dfs_radars[p].rp_pritype;
				ft->ft_rssithresh = dfs_radars[p].rp_rssithresh;
				ft->ft_rssimargin = dfs_radars[p].rp_rssimargin;
				/* Adjust the filter threshold for rssi in non TURBO mode*/
				if( ! (sc->sc_curchan.channelFlags & CHANNEL_TURBO) ) {
					ft->ft_rssithresh += ft->ft_rssimargin;
				}
				ft->ft_minpri = 1000000;

				if (ft->ft_rssithresh < min_rssithresh)
					min_rssithresh = ft->ft_rssithresh;
				if (ft->ft_maxdur > max_pulsedur) 
					max_pulsedur = ft->ft_maxdur;
				for (i=ft->ft_mindur; i<=ft->ft_maxdur; i++) {
					u_int32_t stop=0,tableindex=0;
					while ((tableindex < DFS_MAX_RADAR_OVERLAP) && (!stop)) {
						if ((dfs->dfs_radartable[i])[tableindex] == -1)
							stop = 1;
						else
							tableindex++;
					}
					if (stop) {
						(dfs->dfs_radartable[i])[tableindex] = 
							(int8_t) (dfs->dfs_rinfo.rn_numradars);
					} else {
						DPRINTF(sc, ATH_DEBUG_DFS,
							"%s: Too many overlapping radar filters\n",
							__func__);
						goto bad4;
					}
				}
				dfs->dfs_rinfo.rn_numradars++;
			}
			rf = &(ft->ft_filters[ft->ft_numfilters++]);
			dfs_reset_delayline(&rf->rf_dl);
			
			T = (100000000/dfs_radars[p].rp_pulsefreq) -
				(100*(dfs_radars[p].rp_meanoffset));

			rf->rf_numpulses = dfs_radars[p].rp_numpulses;
			rf->rf_pritype = dfs_radars[p].rp_pritype;
			rf->rf_pulseid = dfs_radars[p].rp_pulseid;
			rf->rf_mindur = dfs_radars[p].rp_mindur;
			rf->rf_maxdur = dfs_radars[p].rp_maxdur;
			rf->rf_minpri =
				_round((int32_t)T - (100*(dfs_radars[p].rp_pulsevar)));
			rf->rf_maxpri =
				_round((int32_t)T + (100*(dfs_radars[p].rp_pulsevar)));
			if( rf->rf_minpri < ft->ft_minpri )
				ft->ft_minpri = rf->rf_minpri;
			rf->rf_threshold = dfs_radars[p].rp_threshold;
			
			rf->rf_filterlen = rf->rf_maxpri * rf->rf_numpulses;
			DPRINTF(sc, ATH_DEBUG_DFS,
				"minpri = %d maxpri = %d filterlen = %d filterID = %d\n",
				rf->rf_minpri, rf->rf_maxpri, rf->rf_filterlen, rf->rf_pulseid);
		}

		dfs->dfs_rinfo.rn_numbin5radars  = numb5radars;
		dfs->dfs_b5radars = kmalloc(numb5radars * sizeof(struct dfs_bin5radars),
					    GFP_KERNEL);
		if (dfs->dfs_b5radars == NULL) {
			DPRINTF(sc, ATH_DEBUG_DFS, "%s: cannot allocate memory for bin5 radars\n",
				__func__);
			goto bad4;
		}
		for (n=0; n<numb5radars; n++) {
			dfs->dfs_b5radars[n].br_pulse = b5pulses[n];
			dfs->dfs_b5radars[n].br_pulse.b5_timewindow *= 1000000;
			if (dfs->dfs_b5radars[n].br_pulse.b5_rssithresh < min_rssithresh)
				min_rssithresh = dfs->dfs_b5radars[n].br_pulse.b5_rssithresh;
			if (dfs->dfs_b5radars[n].br_pulse.b5_maxdur > max_pulsedur )
				max_pulsedur = dfs->dfs_b5radars[n].br_pulse.b5_maxdur;

		}
		dfs_reset_alldelaylines(sc);
		dfs_reset_radarq(sc);
	}
	dfs->dfs_curchan_radindex = -1;
	dfs->dfs_rinfo.rn_minrssithresh = min_rssithresh;
	/* Convert durations to TSF ticks */
	dfs->dfs_rinfo.rn_maxpulsedur = _round((int32_t)((max_pulsedur*100/80)*100));
	DPRINTF (sc, ATH_DEBUG_DFS, "DFS min filter rssiThresh = %d\n",min_rssithresh);
	DPRINTF (sc, ATH_DEBUG_DFS, "DFS max pulse dur = %d ticks\n", dfs->dfs_rinfo.rn_maxpulsedur);
	return 0;
bad4:
	for (n=0; n<256; n++) {
		if (dfs->dfs_radartable[n] != NULL) {
			kfree(dfs->dfs_radartable[n]);
			dfs->dfs_radartable[n] = NULL;
		}
	}
bad2:
	kfree(dfs->dfs_radartable);
	dfs->dfs_radartable = NULL;
bad1:
	for (n=0; n<DFS_MAX_RADAR_TYPES; n++) {
		if (dfs->dfs_radarf[n] != NULL) {
			kfree(dfs->dfs_radarf[n]);
			dfs->dfs_radarf[n] = NULL;
		}
	}
	if (sc->sc_dfs) {
		kfree(sc->sc_dfs);
		sc->sc_dfs = NULL;
	}
	return 1;
#undef N
#else
	if (ath_hal_getcapability(sc->sc_ah, HAL_CAP_PHYDIAG, HAL_CAP_AR, NULL) == HAL_OK) {
		dfs->dfs_proc_phyerr |= DFS_AR_EN;
		dfs_reset_ar(sc);
		dfs_reset_arq(sc);
	}
	dfs->dfs_curchan_radindex = -1;
#endif /* ATH_SUPPORT_DFS */
	return 0;
}
EXPORT_SYMBOL(dfs_attach);

void
dfs_detach(struct ath_softc *sc)
{
	struct ath_dfs *dfs = sc->sc_dfs;
#ifdef ATH_SUPPORT_DFS
	int n, empty;
#endif

	/* Unregister the module here... XXX Still have to figure out how to do that */
	if (dfs == NULL) {
	        DPRINTF(sc, ATH_DEBUG_DFS, "%s: sc_dfs is NULL\n",
			__func__);
		return;
	}
#ifdef ATH_SUPPORT_DFS
	ATH_DFSQ_LOCK(dfs);
	empty = STAILQ_EMPTY(&(dfs->dfs_radarq));
	ATH_DFSQ_UNLOCK(dfs);
	if (!empty) {
		dfs_reset_alldelaylines(sc);
		if (dfs->dfs_b5radars != NULL) {
			kfree(dfs->dfs_b5radars);
			dfs->dfs_b5radars=NULL;
		}
		if (dfs->dfs_nol != NULL) {
			struct dfs_nolelem *nol, *next;
			nol = dfs->dfs_nol;
			while (nol != NULL) {
				next = nol->nol_next;
				kfree(nol);
				nol = next;
			}
			dfs->dfs_nol = NULL;
		}
	}
	for (n=0; n<DFS_MAX_RADAR_TYPES;n++) {
		if (dfs->dfs_radarf[n] != NULL) {
			kfree(dfs->dfs_radarf[n]);
			dfs->dfs_radarf[n] = NULL;
		}
	}
	if (dfs->dfs_radartable != NULL) {
		for (n=0; n<256; n++) {
			if (dfs->dfs_radartable[n] != NULL) {
				kfree(dfs->dfs_radartable[n]);
				dfs->dfs_radartable[n] = NULL;
			}
		}
		kfree(dfs->dfs_radartable);
		dfs->dfs_radartable = NULL;
		sc->sc_ic.ic_isdfsregdomain = 0;
	}
#endif /* ATH_SUPPORT_DFS */
	dfs_reset_ar(sc);

	ATH_ARQ_LOCK(dfs);
	empty = STAILQ_EMPTY(&(dfs->dfs_arq));
	ATH_ARQ_UNLOCK(dfs);
	if (!empty) {
		dfs_reset_arq(sc);
	}
	kfree(STAILQ_FIRST(&(dfs->dfs_eventq)));
	kfree(dfs);
}
EXPORT_SYMBOL(dfs_detach);

static void
dfs_reset_ar(struct ath_softc *sc)
{
	struct ath_dfs *dfs=sc->sc_dfs;

	if (dfs == NULL) {
		DPRINTF(sc,ATH_DEBUG_DFS, "%s: sc_dfs is NULL\n",
			__func__);
		return;
	}
	memset(&dfs->dfs_ar_state, 0, sizeof(dfs->dfs_ar_state));
	dfs->dfs_ar_state.ar_packetthreshold = DFS_AR_PKT_COUNT_THRESH;
	dfs->dfs_ar_state.ar_parthreshold = DFS_AR_ACK_DETECT_PAR_THRESH;
}

static void
dfs_reset_arq(struct ath_softc *sc)
{
	struct ath_dfs *dfs = sc->sc_dfs;
	struct dfs_event *event;

	if (dfs == NULL) {
		DPRINTF(sc,ATH_DEBUG_DFS, "%s: sc_dfs is NULL\n",
			__func__);
		return;
	}
	ATH_ARQ_LOCK(dfs);
	ATH_DFSEVENTQ_LOCK(dfs);
	while (!STAILQ_EMPTY(&(dfs->dfs_arq))) {
		event = STAILQ_FIRST(&(dfs->dfs_arq));
		STAILQ_REMOVE_HEAD(&(dfs->dfs_arq), re_list);
		memset(event, 0, sizeof(struct dfs_event));
		STAILQ_INSERT_TAIL(&(dfs->dfs_eventq), event, re_list);
	}
	ATH_DFSEVENTQ_UNLOCK(dfs);
	ATH_ARQ_UNLOCK(dfs);
}

void
ath_ar_enable(struct ath_softc *sc)
{
	struct ath_dfs *dfs=sc->sc_dfs;
	struct dfs_ar_state *ar;
	HAL_CHANNEL *chan= &sc->sc_curchan;
	HAL_PHYERR_PARAM pe;
	u_int32_t rfilt=0;

	if (dfs == NULL) {
		DPRINTF(sc, ATH_DEBUG_DFS, "%s: sc_dfs is NULL\n",
			__func__);
		return;
	}

	ar = (struct dfs_ar_state *) &dfs->dfs_ar_state;
	if (dfs->dfs_proc_phyerr & (DFS_RADAR_EN | DFS_AR_EN)) {
		if ((chan->channelFlags & CHANNEL_108G) == CHANNEL_108G) {
			/* We are in turbo G, so enable AR */
			DPRINTF(sc, ATH_DEBUG_DFS, "%s: Enabling AR\n", __func__);
			dfs_reset_ar(sc);
			ar->ar_radarrssi = DFS_AR_RADAR_RSSI_THR;
			rfilt = ath_hal_getrxfilter(sc->sc_ah);
			pe.pe_firpwr = HAL_PHYERR_PARAM_NOVAL;
			pe.pe_height = HAL_PHYERR_PARAM_NOVAL;
			pe.pe_prssi = HAL_PHYERR_PARAM_NOVAL;
			pe.pe_inband = HAL_PHYERR_PARAM_NOVAL;
			pe.pe_relpwr = HAL_PHYERR_PARAM_NOVAL;
			pe.pe_maxlen = HAL_PHYERR_PARAM_NOVAL;
			pe.pe_relstep = HAL_PHYERR_PARAM_NOVAL;
			pe.pe_enrelpwr = 0;
			pe.pe_enrelstep = 0;
			pe.pe_enusefir128 = 0;
			pe.pe_enblock = 0;
			pe.pe_enmaxrssi = 0;
			pe.pe_rrssi = ar->ar_radarrssi;
			ath_hal_enabledfs(sc->sc_ah, &pe);
			/* Enable strong signal fast antenna diversity since accurate 
			 * 1-2us radar
			 * detection is not important for AR anyways.
			 */
			ath_hal_setcapability(sc->sc_ah, HAL_CAP_DIVERSITY, 
					      HAL_CAP_STRONG_DIV, dfs->dfs_rinfo.rn_fastdivGCval, NULL);
			rfilt |= HAL_RX_FILTER_PHYRADAR;
			ath_hal_setrxfilter(sc->sc_ah, rfilt);
		}
	} else {
		DPRINTF(sc, ATH_DEBUG_DFS, "%s: Disabling AR\n", __func__);
		rfilt = ath_hal_getrxfilter(sc->sc_ah);
		rfilt &= ~HAL_RX_FILTER_PHYRADAR;
		ath_hal_setrxfilter(sc->sc_ah,rfilt);
		/* Enable strong signal fast antenna diversity */
		ath_hal_setcapability(sc->sc_ah, HAL_CAP_DIVERSITY, 
				      HAL_CAP_STRONG_DIV, dfs->dfs_rinfo.rn_fastdivGCval, NULL);

	}
}
EXPORT_SYMBOL(ath_ar_enable);

int dfs_radar_enable(struct ath_softc *sc)
{
	struct ath_dfs *dfs=sc->sc_dfs;
	u_int32_t rfilt;
	HAL_CHANNEL *chan=&sc->sc_curchan;
	struct ath_hal *ah = sc->sc_ah;

	if (dfs == NULL) {
		DPRINTF(sc, ATH_DEBUG_DFS, "%s: sc_dfs is NULL\n",
			__func__);
		return -EIO;
	}
	rfilt = ath_hal_getrxfilter(ah);
	if ((dfs->dfs_proc_phyerr & (DFS_RADAR_EN | DFS_AR_EN)) &&
	    (sc->sc_ic.ic_opmode == IEEE80211_M_HOSTAP)) {
#ifdef ATH_SUPPORT_DFS
		if (chan->channelFlags & CHANNEL_DFS) {
			struct dfs_state *rs;
			u_int8_t index;
			HAL_PHYERR_PARAM pe;

			/*
			 * Disable radar detection in case we need to setup
			 * a new channel state and radars are somehow being
			 * reported. Avoids locking problem.
			 */

			rfilt = ath_hal_getrxfilter(ah);
			rfilt &= ~HAL_RX_FILTER_PHYRADAR;
			ath_hal_setrxfilter(ah, rfilt);
			/* Enable strong signal fast antenna diversity */
			ath_hal_setcapability(ah, HAL_CAP_DIVERSITY, 
					      HAL_CAP_STRONG_DIV, dfs->dfs_rinfo.rn_fastdivGCval, NULL);
			dfs_reset_alldelaylines(sc);
			
			rs = dfs_getchanstate(sc, &index);
			if (rs != NULL) {
				if (index != dfs->dfs_curchan_radindex)
					dfs_reset_alldelaylines(sc);
				dfs->dfs_curchan_radindex = (int16_t) index;

				pe.pe_firpwr = rs->rs_firpwr;
				pe.pe_rrssi = rs->rs_radarrssi;
				pe.pe_height = rs->rs_height;
				pe.pe_prssi = rs->rs_pulserssi;
				pe.pe_inband = rs->rs_inband;
				/* 5413 specific */
				pe.pe_relpwr = rs->rs_relpwr;
				pe.pe_relstep = rs->rs_relstep;
				pe.pe_maxlen = rs->rs_maxlen;


				/* Disable strong signal fast antenna diversity */
				ath_hal_setcapability(ah, HAL_CAP_DIVERSITY,
						      HAL_CAP_STRONG_DIV, 1, NULL);

				ath_hal_enabledfs(sc->sc_ah, &pe);
				rfilt |= HAL_RX_FILTER_PHYRADAR;
				ath_hal_setrxfilter(ah, rfilt);
				DPRINTF(sc, ATH_DEBUG_DFS, "Enabled radar detection on channel %d\n",
					chan->channel);

				/*
				 * Fast antenna diversity for strong signals disturbs
				 * radar detection of 1-2us pusles. Enable fast diveristy
				 * but disable the strong signal aspect of it
				 */
				if (ath_hal_getcapability(ah, HAL_CAP_DIVERSITY, 1, NULL) == HAL_OK) {
					ath_hal_setcapability(ah, HAL_CAP_DIVERSITY, 
							      HAL_CAP_STRONG_DIV, 0, NULL);
				}
			} else
				DPRINTF(sc, ATH_DEBUG_DFS, "%s: No more radar states left\n",
					__func__);
		} else {
			if (!(chan->channelFlags & CHANNEL_2GHZ)) {
				/* Enable strong signal fast antenna diversity */
				ath_hal_setcapability(ah, HAL_CAP_DIVERSITY, 
						      HAL_CAP_STRONG_DIV, dfs->dfs_rinfo.rn_fastdivGCval, NULL);
				/* Disable Radar if not 2GHZ channel and not DFS */
				rfilt &= ~HAL_RX_FILTER_PHYRADAR;
				ath_hal_setrxfilter(ah, rfilt);
			}
		}
#endif
	} else {
		/* Enable strong signal fast antenna diversity */
		ath_hal_setcapability(ah, HAL_CAP_DIVERSITY, 
				      HAL_CAP_STRONG_DIV, dfs->dfs_rinfo.rn_fastdivGCval, NULL);
		/* Disable Radar if RADAR or AR not enabled */
		rfilt &= ~HAL_RX_FILTER_PHYRADAR;
		ath_hal_setrxfilter(ah, rfilt);
	}
	return 0;
}
EXPORT_SYMBOL(dfs_radar_enable);

/*
 * Disable AR if AR channel
 */

void
ath_ar_disable(struct ath_softc *sc)
{
	u_int32_t rfilt;
	HAL_CHANNEL *chan=&sc->sc_curchan;
	struct ath_dfs *dfs=sc->sc_dfs;
	
	if (dfs == NULL) {
		DPRINTF(sc, ATH_DEBUG_DFS, "%s: sc_dfs is NULL\n",
			__func__);
		return;
	}
	if ((chan->channelFlags & CHANNEL_108G) == CHANNEL_108G) {
		/* Enable strong signal fast antenna diversity */
		ath_hal_setcapability(sc->sc_ah, HAL_CAP_DIVERSITY, 
				      HAL_CAP_STRONG_DIV, dfs->dfs_rinfo.rn_fastdivGCval, NULL);
		rfilt = ath_hal_getrxfilter(sc->sc_ah);
		rfilt &= ~HAL_RX_FILTER_PHYERR;
		ath_hal_setrxfilter(sc->sc_ah, rfilt);
		dfs_reset_ar(sc);
		dfs_reset_arq(sc);
	}
}
EXPORT_SYMBOL(ath_ar_disable);

void
ath_process_phyerr(struct ath_softc *sc, struct ath_desc *ds, u_int64_t fulltsf)
{
	struct ath_dfs *dfs=sc->sc_dfs;
	HAL_CHANNEL *chan=&sc->sc_curchan;
	struct dfs_event *event;
	u_int8_t rssi;
	u_int32_t dur;
	int empty;

	if (ds->ds_rxstat->rs_phyerr != HAL_PHYERR_RADAR)
		return;

	if (dfs == NULL) {
		DPRINTF(sc, ATH_DEBUG_DFS, "%s: sc_dfs is NULL\n",__func__);
		return;
	}

	rssi = (u_int8_t) ds->ds_rxstat->rs_rssi;
	/* Only lower 8 bits of dur are valid */
	dur = (ds->ds_rxstat->rs_datalen && ds->ds_vdata != NULL ?
	       (u_int32_t)(*((u_int8_t *) ds->ds_vdata)) : 0) & 0xff;
	if ((rssi == 0) && (dur == 0))
		return;
	ATH_DFSEVENTQ_LOCK(dfs);
	empty = STAILQ_EMPTY(&(dfs->dfs_eventq));
	ATH_DFSEVENTQ_UNLOCK(dfs);
	if (empty)
		return;
	if ((chan->channelFlags & CHANNEL_108G) == CHANNEL_108G) {
		ATH_DFSEVENTQ_LOCK(dfs);
		event = STAILQ_FIRST(&(dfs->dfs_eventq));
		if (event == NULL) {
			ATH_DFSEVENTQ_UNLOCK(dfs);
			DPRINTF(sc, ATH_DEBUG_DFS, "%s: no more events space left\n",
				__func__);
			return;
		}
		STAILQ_REMOVE_HEAD(&(dfs->dfs_eventq), re_list);
		ATH_DFSEVENTQ_UNLOCK(dfs);
		
		event->re_rssi = rssi;
		event->re_dur = dur;
		event->re_full_ts = fulltsf;
		event->re_ts = (ds->ds_rxstat->rs_tstamp) & DFS_TSMASK;
		event->re_chanindex = dfs->dfs_curchan_radindex;

		ATH_ARQ_LOCK(dfs);
		STAILQ_INSERT_TAIL(&(dfs->dfs_arq), event, re_list);
		ATH_ARQ_UNLOCK(dfs);
	}
#ifdef ATH_SUPPORT_DFS
	else
		if (chan->channelFlags & CHANNEL_DFS) {
			if (rssi < dfs->dfs_rinfo.rn_minrssithresh ||
				dur > dfs->dfs_rinfo.rn_maxpulsedur )
				return;
			ATH_DFSEVENTQ_LOCK(dfs);
			event = STAILQ_FIRST(&(dfs->dfs_eventq));
			if (event == NULL) {
				ATH_DFSEVENTQ_UNLOCK(dfs);
				DPRINTF(sc, ATH_DEBUG_DFS, "%s: no more events space left\n",
					__func__);
				return;
			}
			STAILQ_REMOVE_HEAD(&(dfs->dfs_eventq), re_list);
			ATH_DFSEVENTQ_UNLOCK(dfs);

			event->re_rssi = rssi;
			event->re_dur = dur;
			event->re_full_ts = fulltsf;
			event->re_ts = (ds->ds_rxstat->rs_tstamp) & DFS_TSMASK;
			event->re_chanindex = dfs->dfs_curchan_radindex;

			ATH_DFSQ_LOCK(dfs);
			STAILQ_INSERT_TAIL(&(dfs->dfs_radarq), event, re_list);
			ATH_DFSQ_UNLOCK(dfs);
		}
#endif /* ATH_SUPPORT_DFS */
}
EXPORT_SYMBOL(ath_process_phyerr);
		
#define UPDATE_TOP_THREE_PEAKS(_histo, _peakPtrList, _currWidth) { \
	if ((_histo)[(_peakPtrList)[0]] < (_histo)[(_currWidth)]) {	\
		(_peakPtrList)[2] = (_currWidth != (_peakPtrList)[1]) ?	\
					(_peakPtrList)[1] : (_peakPtrList)[2];  \
		(_peakPtrList)[1] = (_peakPtrList)[0]; \
		(_peakPtrList)[0] = (_currWidth); \
	} else if ((_currWidth != (_peakPtrList)[0])	\
			&& ((_histo)[(_peakPtrList)[1]] < (_histo)[(_currWidth)])) { \
		(_peakPtrList)[2] = (_peakPtrList)[1]; \
		(_peakPtrList)[1] = (_currWidth);      \
	} else if ((_currWidth != (_peakPtrList)[1])   \
			&& (_currWidth != (_peakPtrList)[0])  \
			&& ((_histo)[(_peakPtrList)[2]] < (_histo)[(_currWidth)])) { \
		(_peakPtrList)[2] = (_currWidth);  \
	} \
}                                                                           

/*
 * This routine builds the histogram based on radar duration and does pattern matching
 * on incoming radars to determine if neighboring traffic is present.
 */

void
dfs_process_ar_event(struct ath_softc *sc, HAL_CHANNEL *chan)
{
	struct ath_dfs *dfs=sc->sc_dfs;
	struct dfs_ar_state *ar;
	struct dfs_event *re=NULL;
	u_int32_t sumpeak=0,numpeaks,rssi,width,origregionsum=0, i;
	u_int16_t thistimestamp;
	int empty;

	if (dfs == NULL) {
		DPRINTF(sc, ATH_DEBUG_DFS, "%s: sc_dfs is NULL\n",
			__func__);
		return;
	}
	ar = (struct dfs_ar_state *) &(dfs->dfs_ar_state);
	ATH_ARQ_LOCK(dfs);
	empty = STAILQ_EMPTY(&(dfs->dfs_arq));
	ATH_ARQ_UNLOCK(dfs);
	while (!empty) {
		ATH_ARQ_LOCK(dfs);
		re = STAILQ_FIRST(&(dfs->dfs_arq));
		if (re != NULL)
			STAILQ_REMOVE_HEAD(&(dfs->dfs_arq), re_list);
		ATH_ARQ_UNLOCK(dfs);
		if (re == NULL)
			return;

		thistimestamp = re->re_ts;
		rssi = re->re_rssi;
		width = re->re_dur;

		/* Return the dfs event to the free event list */
		memset(re, 0, sizeof(struct dfs_event));
		ATH_DFSEVENTQ_LOCK(dfs);
		STAILQ_INSERT_TAIL(&(dfs->dfs_eventq), re, re_list);
		ATH_DFSEVENTQ_UNLOCK(dfs);

		/* determine if current radar is an extension of previous radar */
		if (ar->ar_prevwidth == 255) {
			/* tag on previous width for consideraion of low data rate ACKs */
			ar->ar_prevwidth += width;
			width = (width == 255) ? 255 : ar->ar_prevwidth;
		} else if ((width == 255) &&
			   (ar->ar_prevwidth == 510 ||
			    ar->ar_prevwidth == 765 ||
			    ar->ar_prevwidth == 1020)) {
			/* Aggregate up to 5 consecuate max radar widths
			 * to consider 11Mbps long preamble 1500-byte pkts
			 */
			ar->ar_prevwidth += width;
		} else if (ar->ar_prevwidth == 1275 && width != 255) {
			/* Found 5th consecute maxed out radar, reset history */
			width += ar->ar_prevwidth;
			ar->ar_prevwidth = 0;
		} else if (ar->ar_prevwidth > 255) {
			/* Ignore if there are less than 5 consecutive maxed out radars */
			ar->ar_prevwidth = width;
			width = 255;
		} else {
			ar->ar_prevwidth = width;
		}
		/* For ignoring noises with radar duration in ranges of 3-30: AP4x */
		if ((width >= 257 && width <= 278) ||	/* Region 7 - 5.5Mbps (long pre) ACK = 270 = 216 us */
		    (width >= 295 && width <= 325) ||	/* Region 8 - 2Mbps (long pre) ACKC = 320 = 256us */
		    (width >= 1280 && width <= 1300)) {
			u_int16_t wraparoundadj=0;
			u_int16_t base = (width >= 1280) ? 1275 : 255;
			if (thistimestamp < ar->ar_prevtimestamp) {
				wraparoundadj = 32768;
			}
			if ((thistimestamp + wraparoundadj - ar->ar_prevtimestamp) !=
			    (width - base)) {
				width = 1;
			}
		}
		if (width <= 10) {
			ATH_ARQ_LOCK(dfs);
			empty = STAILQ_EMPTY(&(dfs->dfs_arq));
			ATH_ARQ_UNLOCK(dfs);
			continue;
		}
		
		/*
		 * Overloading the width=2 in: Store a count of radars w/max duration
		 * and high RSSI (not noise)
		 */
		if ((width == 255) && (rssi > DFS_AR_RSSI_THRESH_STRONG_PKTS))
			width = 2;
		
		/*
		 * Overloading the width=3 bin:
		 *   Double and store a count of rdars of durtaion that matches 11Mbps (long preamble)
		 *   TCP ACKs or 1500-byte data packets
		 */
		if ((width >= 1280 && width <= 1300) ||
		    (width >= 318 && width <= 325)) {
			width = 3;
			ar->ar_phyerrcount[3] += 2;
			ar->ar_acksum += 2;
		}
		
		/* build histogram of radar duration */
		if (width > 0 && width <= 510)
			ar->ar_phyerrcount[width]++;
		else {
			/* invalid radar width, throw it away */
			ATH_ARQ_LOCK(dfs);
			empty = STAILQ_EMPTY(&(dfs->dfs_arq));
			ATH_ARQ_UNLOCK(dfs);
			continue;
		}
		/* Received radar of interest (i.e., signature match), proceed to check if
		 * there is enough neighboring traffic to drop out of Turbo 
		 */
		if ((width >= 33 && width <= 38) ||          /* Region 0: 24Mbps ACK = 35 = 28us */
		    (width >= 39 && width <= 44) ||          /* Region 1: 12Mbps ACK = 40 = 32us */
		    (width >= 53 && width <= 58) ||          /* Region 2:  6Mbps ACK = 55 = 44us */
		    (width >= 126 && width <= 140) ||        /* Region 3: 11Mbps ACK = 135 = 108us */
		    (width >= 141 && width <= 160) ||        /* Region 4: 5.5Mbps ACK = 150 = 120us */
		    (width >= 189 && width <= 210) ||        /* Region 5:  2Mbps ACK = 200 = 160us */
		    (width >= 360 && width <= 380) ||        /* Region 6   1Mbps ACK = 400 = 320us */
		    (width >= 257 && width <= 270) ||        /* Region 7   5.5Mbps (Long Pre) ACK = 270 = 216us */
		    (width >= 295 && width <= 302) ||        /* Region 8   2Mbps (Long Pre) ACK = 320 = 256us */
		    /* Ignoring Region 9 due to overlap with 255 which is same as board noise */
		    /* Region 9  11Mbps (Long Pre) ACK = 255 = 204us */            
		    (width == 3)) {
			ar->ar_acksum++;
			
			/* double the count for strong radars that match one of the ACK signatures */
			if (rssi > DFS_AR_RSSI_DOUBLE_THRESHOLD) {
				ar->ar_phyerrcount[width]++;
				ar->ar_acksum++;
			}
			UPDATE_TOP_THREE_PEAKS(ar->ar_phyerrcount,
					       ar->ar_peaklist, width);
			/* sum the counts of these peaks */
			numpeaks = DFS_AR_MAX_NUM_PEAKS;
			origregionsum = ar->ar_acksum;
			for (i=0; i<= DFS_AR_MAX_NUM_PEAKS; i++) {
				if (ar->ar_peaklist[i] > 0) {
					if ((i==0) &&
					    (ar->ar_peaklist[i] == 3) &&
					    (ar->ar_phyerrcount[3] <
					     ar->ar_phyerrcount[2]) &&
					    (ar->ar_phyerrcount[3] > 6)) {
						/*
						 * If the top peak is one that
						 * maches the 11Mbps long
						 * preamble TCP Ack/1500-byte
						 * data, include the count for
						 * radars that hav emax
						 * duration and high rssi
						 * (width = 2) to boost the
						 * sum for the PAR test that
						 * follows */
						sumpeak += (ar->ar_phyerrcount[2]
							    + ar->ar_phyerrcount[3]);
						ar->ar_acksum += (ar->ar_phyerrcount[2]
								  + ar->ar_phyerrcount[3]);
					} else {
						sumpeak += ar->ar_phyerrcount[ar->ar_peaklist[i]];
					}
				} else 
					numpeaks--;
			}
			
			/*
			 * If sum of patterns matches exceeds packet threshold,
			 * perform comparison between peak-to-avg ratio against parThreshold
			 */
			if ((ar->ar_acksum > ar->ar_packetthreshold) &&
			    ((sumpeak * DFS_AR_REGION_WIDTH) > (ar->ar_parthreshold * numpeaks *
								ar->ar_acksum))) {
				/* neighboring traffic detected, get out of Turbo */
				chan->privFlags |= CHANNEL_INTERFERENCE;
				
				memset(ar->ar_peaklist, 0, sizeof(ar->ar_peaklist));
				ar->ar_acksum = 0;
				memset(ar->ar_phyerrcount, 0, sizeof(ar->ar_phyerrcount));
			} else {
				/*
				 * reset sum of matches to discount the count of
				 * strong radars with max duration
				 */
				ar->ar_acksum = origregionsum;
			}
		}
		ar->ar_prevtimestamp = thistimestamp;
		ATH_ARQ_LOCK(dfs);
		empty = STAILQ_EMPTY(&(dfs->dfs_arq));
		ATH_ARQ_UNLOCK(dfs);
	}
}
EXPORT_SYMBOL(dfs_process_ar_event);

int
ath_ioctl_phyerr(struct ath_softc *sc, struct ath_diag *ad)
{
	u_int id= ad->ad_id & ATH_DIAG_ID;
	void *indata=NULL;
	void *outdata=NULL;
	u_int32_t insize = ad->ad_in_size;
	u_int32_t outsize = ad->ad_out_size;
	int error=0;
	u_int32_t *data=NULL;
	HAL_PHYERR_PARAM peout;
	struct ath_dfs *dfs=sc->sc_dfs;
	struct dfs_ioctl_params *dfsparams;


	if (dfs == NULL) {
		error = -EINVAL;
		goto bad;
	}
	if (ad->ad_id & ATH_DIAG_IN) {
		/*
		 * Copy in data.
		 */
		indata = kmalloc(insize, GFP_KERNEL);
		if (indata == NULL) {
			error = -ENOMEM;
			goto bad;
		}
		if (copy_from_user(indata, ad->ad_in_data, insize)) {
			error = -EFAULT;
			goto bad;
		}
	}
	if (ad->ad_id & ATH_DIAG_DYN) {
		/*
		 * Allocate a buffer for the results (otherwise the HAL
		 * returns a pointer to a buffer where we can read the
		 * results).  Note that we depend on the HAL leaving this
		 * pointer for us to use below in reclaiming the buffer;
		 * may want to be more defensive.
		 */
		outdata = kmalloc(outsize, GFP_KERNEL);
		if (outdata == NULL) {
			error = -ENOMEM;
			goto bad;
		}
	}
	switch (id) {
	case DFS_MUTE_TIME:
		if (insize < sizeof(u_int32_t)) {
			error = -EINVAL;
			break;
		}
		data = (u_int32_t *) indata;
		sc->sc_dfstesttime = *data;
		break;
	case DFS_SET_THRESH:
		if (insize < sizeof(HAL_PHYERR_PARAM)) {
			error = -EINVAL;
			break;
		}
		dfsparams = (struct dfs_ioctl_params *) indata;
		if (!dfs_set_thresholds(sc, DFS_PARAM_FIRPWR, dfsparams->dfs_firpwr))
			error = -EINVAL;
		if (!dfs_set_thresholds(sc, DFS_PARAM_RRSSI, dfsparams->dfs_rrssi))
			error = -EINVAL;
		if (!dfs_set_thresholds(sc, DFS_PARAM_HEIGHT, dfsparams->dfs_height))
			error = -EINVAL;
		if (!dfs_set_thresholds(sc, DFS_PARAM_PRSSI, dfsparams->dfs_prssi))
			error = -EINVAL;
		if (!dfs_set_thresholds(sc, DFS_PARAM_INBAND, dfsparams->dfs_inband))
			error = -EINVAL;
		/* 5413 speicfic */
		if (!dfs_set_thresholds(sc, DFS_PARAM_ENRELPWR, dfsparams->dfs_enrelpwr))
			error = -EINVAL;
		if (!dfs_set_thresholds(sc, DFS_PARAM_RELPWR, dfsparams->dfs_relpwr))
			error = -EINVAL;
		if (!dfs_set_thresholds(sc, DFS_PARAM_ENUSEFIR128, dfsparams->dfs_enusefir128))
			error = -EINVAL;
		if (!dfs_set_thresholds(sc, DFS_PARAM_ENBLOCK, dfsparams->dfs_enblock))
			error = -EINVAL;
		if (!dfs_set_thresholds(sc, DFS_PARAM_ENMAXRSSI, dfsparams->dfs_enmaxrssi))
			error = -EINVAL;
		if (!dfs_set_thresholds(sc, DFS_PARAM_ENRELSTEP, dfsparams->dfs_enrelstep))
			error = -EINVAL;
		if (!dfs_set_thresholds(sc, DFS_PARAM_RELSTEP, dfsparams->dfs_relstep))
			error = -EINVAL;
		if (!dfs_set_thresholds(sc, DFS_PARAM_MAXLEN, dfsparams->dfs_maxlen))
			error = -EINVAL;
		break;
	case DFS_GET_THRESH:
		outsize = sizeof(struct dfs_ioctl_params);
		ath_hal_getdfsthresh(sc->sc_ah, &peout);
		dfsparams = (struct dfs_ioctl_params *) outdata;
		dfsparams->dfs_firpwr = peout.pe_firpwr;
		dfsparams->dfs_rrssi = peout.pe_rrssi;
		dfsparams->dfs_height = peout.pe_height;
		dfsparams->dfs_prssi = peout.pe_prssi;
		dfsparams->dfs_inband = peout.pe_inband;
		/* 5413 speicfic */
		dfsparams->dfs_enrelpwr = peout.pe_enrelpwr;
		dfsparams->dfs_relpwr = peout.pe_relpwr;
		dfsparams->dfs_enusefir128 = peout.pe_enusefir128;
		dfsparams->dfs_enblock = peout.pe_enblock;
		dfsparams->dfs_enmaxrssi = peout.pe_enmaxrssi;
		dfsparams->dfs_enrelstep = peout.pe_enrelstep;
		dfsparams->dfs_relstep = peout.pe_relstep;
		dfsparams->dfs_maxlen = peout.pe_maxlen;
		break;
	case DFS_GET_USENOL:
		outsize = sizeof(u_int32_t);
		*((u_int32_t *)outdata) = dfs->dfs_rinfo.rn_use_nol;
		break;
	case DFS_SET_USENOL:
                if (insize < sizeof(u_int32_t)) {
                        error = -EINVAL;
                        break;
                }
		dfs->dfs_rinfo.rn_use_nol = *(u_int32_t *)indata;
		// iwpriv markdfs can do the same thing...
		if ( dfs->dfs_rinfo.rn_use_nol )
			sc->sc_ic.ic_flags_ext |= IEEE80211_FEXT_MARKDFS;
		else
			sc->sc_ic.ic_flags_ext &= ~IEEE80211_FEXT_MARKDFS;
		break;
	case DFS_BANGRADAR:
		sc->sc_ic.ic_bangradar = 1;     
		break;
        case DFS_RADARDETECTS:
		outsize = sizeof (u_int32_t);
		*((u_int32_t *)outdata) = dfs->stats.num_radar_detects;
                break;
	default:
		error = -EINVAL;
	}
	if (outsize < ad->ad_out_size)
		ad->ad_out_size = outsize;
	if (outdata && 
	    copy_to_user(ad->ad_out_data, outdata, ad->ad_out_size))
		error = -EFAULT;
bad:
	if ((ad->ad_id & ATH_DIAG_IN) && indata != NULL)
		kfree(indata);
	if ((ad->ad_id & ATH_DIAG_DYN) && outdata != NULL)
		kfree(outdata);
	return error;
}
EXPORT_SYMBOL(ath_ioctl_phyerr);


#if defined(ATH_SUPPORT_DFS)
static int
dfs_set_thresholds(struct ath_softc *sc, const u_int32_t threshtype,
		   const u_int32_t value)
{
	struct ath_dfs *dfs=sc->sc_dfs;
	int16_t chanindex;
	struct dfs_state *rs;
	HAL_PHYERR_PARAM pe;

	if (dfs == NULL) {
		DPRINTF(sc, ATH_DEBUG_DFS, "%s: sc_dfs is NULL\n",
			__func__);
		return 0;
	}

	chanindex = dfs->dfs_curchan_radindex;
	if ((chanindex <0) || (chanindex >= DFS_NUM_RADAR_STATES))
		return 0;
	pe.pe_firpwr = HAL_PHYERR_PARAM_NOVAL;
	pe.pe_height = HAL_PHYERR_PARAM_NOVAL;
	pe.pe_prssi = HAL_PHYERR_PARAM_NOVAL;
	pe.pe_inband = HAL_PHYERR_PARAM_NOVAL;
	pe.pe_rrssi = HAL_PHYERR_PARAM_NOVAL;
	/* 5413 specific */
	pe.pe_enrelpwr = HAL_PHYERR_PARAM_NOVAL;
	pe.pe_relpwr = HAL_PHYERR_PARAM_NOVAL;
	pe.pe_enusefir128 = HAL_PHYERR_PARAM_NOVAL;
	pe.pe_enblock = HAL_PHYERR_PARAM_NOVAL;
	pe.pe_enmaxrssi = HAL_PHYERR_PARAM_NOVAL;
	pe.pe_enrelstep = HAL_PHYERR_PARAM_NOVAL;
	pe.pe_relstep = HAL_PHYERR_PARAM_NOVAL;
	pe.pe_maxlen = HAL_PHYERR_PARAM_NOVAL;

	rs = &(dfs->dfs_radar[chanindex]);
	switch (threshtype) {
	case DFS_PARAM_FIRPWR:
		rs->rs_firpwr = (int32_t) value;
		pe.pe_firpwr = rs->rs_firpwr;
		break;
	case DFS_PARAM_RRSSI:
		rs->rs_radarrssi = value;
		pe.pe_rrssi = value;
		break;
	case DFS_PARAM_HEIGHT:
		rs->rs_height = value;
		pe.pe_height = value;
		break;
	case DFS_PARAM_PRSSI:
		rs->rs_pulserssi = value;
		pe.pe_prssi = value;
		break;
	case DFS_PARAM_INBAND:
		rs->rs_inband = value;
		pe.pe_inband = value;
		break;
	/* 5413 specific */
	case DFS_PARAM_ENRELPWR:
		rs->rs_enrelpwr = value;
		pe.pe_enrelpwr = value;
		break;
	case DFS_PARAM_RELPWR:
		rs->rs_relpwr = value;
		pe.pe_relpwr = value;
		break;
	case DFS_PARAM_ENUSEFIR128:
		rs->rs_enusefir128 = value;
		pe.pe_enusefir128 = value;
		break;
	case DFS_PARAM_ENBLOCK:
		rs->rs_enblock = value;
		pe.pe_enblock = value;
		break;
	case DFS_PARAM_ENMAXRSSI:
		rs->rs_enmaxrssi = value;
		pe.pe_enmaxrssi = value;
		break;
	case DFS_PARAM_ENRELSTEP:
		rs->rs_enrelstep = value;
		pe.pe_enrelstep = value;
		break;
	case DFS_PARAM_RELSTEP:
		rs->rs_relstep = value;
		pe.pe_relstep = value;
		break;
	case DFS_PARAM_MAXLEN:
		rs->rs_maxlen = value;
		pe.pe_maxlen = value;
		break;
	}
	ath_hal_enabledfs(sc->sc_ah, &pe);
	return 1;
}

int
dfs_get_thresholds(struct ath_softc *sc, HAL_PHYERR_PARAM *param)
{
	ath_hal_getdfsthresh(sc->sc_ah, param);
	return 1;
}
EXPORT_SYMBOL(dfs_get_thresholds);

/*
 * Module glue.
 */
static char *dev_info = "ath_dfs";

MODULE_AUTHOR("Atheros Communications, Inc.");
MODULE_DESCRIPTION("DFS Support for Atheros 802.11 wireless LAN cards.");
MODULE_SUPPORTED_DEVICE("Atheros WLAN cards");
#ifdef MODULE_LICENSE
MODULE_LICENSE("Proprietary");
#endif

static int __init
init_ath_dfs(void)
{
	printk (KERN_INFO "%s: Version 2.0.0\n"
		"Copyright (c) 2005-2006 Atheros Communications, Inc. "
		"All Rights Reserved\n",dev_info);
	return 0;
}
module_init(init_ath_dfs);

static void __exit
exit_ath_dfs(void)
{
	printk (KERN_INFO "%s: driver unloaded\n", dev_info);
}
module_exit(exit_ath_dfs);

#endif /* ATH_SUPPORT_DFS */
